#include <assert.h>
#include <stdlib.h>

#include "tampon.h"

/* invariants de la structure tampon */
#define INVARIANTS_TP(t) \
	(assert((t)     != NULL), \
	 assert((t)->tp != NULL), \
	 assert((t)->pc != NULL), \
	 assert((t)->tll >= 1), \
	 assert((t)->pc - (t)->tp >= 0), \
	 assert((size_t) ((t)->pc - (t)->tp) <= (t)->tll), \
	 assert((t)->tll * TP_FACT_AGR > (t)->tll))

static int tp_plein(char, struct tampon *);

/* initialiser le tampon tp de taille tll */
int
tp_init(struct tampon *tp, size_t tll)
{
	assert(tll >= 1);
	assert(tp != NULL);

	if ((tp->pc = tp->tp = malloc(tll)) == NULL)
		return -1;
	tp->tll = tll;

	INVARIANTS_TP(tp);

	return 0;
}

/* ajouter le caractère c au tampon tp */
int
tp_ecr(char c, struct tampon *tp)
{
	INVARIANTS_TP(tp);

	if ((size_t) (tp->pc - tp->tp) == tp->tll) /* le tampon est-il plein? */
		return tp_plein(c, tp);
	*tp->pc++ = c;

	INVARIANTS_TP(tp);

	return (unsigned char) c;
}

/* agrandir le tampon tp et y ajouter le caractère c */
static int
tp_plein(char c, struct tampon *tp)
{
	size_t diff;

	INVARIANTS_TP(tp);

	diff = tp->pc - tp->tp;
	if ((tp->tp = realloc(tp->tp, tp->tll * TP_FACT_AGR)) == NULL)
		return -1;
	tp->tll  *= TP_FACT_AGR;
	tp->pc    = tp->tp + diff; /* tp->tp pourrait avoir changé */
	*tp->pc++ = c;

	INVARIANTS_TP(tp);

	return (unsigned char) c;
}

/* effacer le contenu du tampon tp */
void
tp_eff(struct tampon *tp)
{
	INVARIANTS_TP(tp);

	tp->pc = tp->tp;

	INVARIANTS_TP(tp);
}

/* effacer les caractères du tampon tp jusqu'à la dernière nouvelle ligne,
 * inclusivement */
void
tp_bloc_rec(struct tampon *tp)
{
	INVARIANTS_TP(tp);

	while (tp->pc > tp->tp) {
		if (*--tp->pc == 'n' && *--tp->pc == '\\')
			break;
	}

	INVARIANTS_TP(tp);
}

/* supprimer le domaine cadet */
int
tp_etq_rec(struct tampon *tp)
{
	INVARIANTS_TP(tp);
	assert(tp->pc - tp->tp > 1);
	assert(*tp->tp == ':');

	while (*--tp->pc != ':')
		;
	if (tp->pc - tp->tp != 0 && *(tp->pc-1) == '\\') {
		--tp->pc;
		assert(tp->pc - tp->tp > 0);
		return ETQ_SEP_FCT;
	}
	assert(tp->pc - tp->tp > 0);
	return ETQ_SEP_REL;
}

/* obtenir le nombre de caractères dans un tampon */
size_t
tp_ncar(struct tampon *tp)
{
	INVARIANTS_TP(tp);

	return tp->pc - tp->tp;
}
